@define-color np_fg_color #ddddde;
@define-color np_bg_color #333;
@define-color np_text_color #2e3436;

GtkWindow {
    color: @np_fg_color;
}

GtkSourceView {
    color: @np_text_color;
    background-color: shade (@np_fg_color, 1.10);
    font: Monospace 10;
}

SushiFontWidget {
    background-color: transparent;
}

.scrollbar.slider,
.scrollbar.button {
    background-color: shade (@np_bg_color, 1.30);
}

.scrollbar.button:prelight,
.scrollbar.slider:prelight {
    background-color: shade (@np_bg_color, 1.45);
}

.scrollbar.button:prelight:active,
.scrollbar.slider:prelight:active {
    background-color: shade (@np_bg_color, 1.70);
}

.np-decoration {
    font: bold;
}

.np-decoration.button {
    background-image: none;
    padding: 4px;

    border-image: none;
    border-style: none;
    border-radius: 0;
    border-width: 0;
    background-color: alpha (#000, 0.0);

    -GtkWidget-focus-line-width: 0;
}

.np-decoration.button:hover {
    background-image: -gtk-gradient(linear,
                                    left top,
                                    left bottom,
                                    from(alpha(@np_bg_color, 0.0)),
                                    color-stop(0.50, alpha(@np_bg_color, 0.66)),
                                    to(alpha(@np_bg_color, 0.0)));
}

.np-decoration.button:hover:active {
    background-image: -gtk-gradient(linear,
                                    left top,
                                    left bottom,
                                    from(alpha(@np_bg_color, 0.0)),
                                    color-stop(0.50, @np_bg_color),
                                    to(alpha(@np_bg_color, 0.0)));
}
